from __future__ import annotations

from typing import Any
from typing import Tuple
from typing import List
from typing import Union
from typing_extensions import Annotated
from nptyping import NDArray

import compas.geometry
import numpy as np

VerticesFacesNumpy = Tuple[
    NDArray[(Any, 3), np.float64],
    NDArray[(Any, 3), np.int32]
]

VerticesFaces = Union[
    VerticesFacesNumpy,
    Tuple[
        List[Annotated[List[float], 3]],
        List[Annotated[List[int], 3]]
    ]
]

ListOfPolylinesNumpy = List[NDArray[(Any, 3), np.float64]]

ListOfPlanes = List[
    Union[
        compas.geometry.Plane,
        Tuple[
            compas.geometry.Point,
            compas.geometry.Vector
        ],
        Tuple[
            Annotated[List[float], 3],
            Annotated[List[float], 3]
        ]
    ]
]



from typing import List
from typing import Tuple
import numpy as np
from compas_cgal._cgal import meshing
from compas.plugins import plugin

from compas_cgal.types import VerticesFaces
from compas_cgal.types import VerticesFacesNumpy


__all__ = [
    'remesh'
]


@plugin(category='trimesh', pluggable_name='trimesh_remesh')
def remesh(mesh: VerticesFaces,
           target_edge_length: float,
           number_of_iterations: int = 10,
           do_project: bool = True) -> VerticesFacesNumpy:
    """Remeshing of a triangle mesh.

    Parameters
    ----------
    mesh : tuple of vertices and faces
        The mesh to remesh.
    target_edge_length : float
        The target edge length.
    number_of_iterations : int, optional
        Number of remeshing iterations.
        Default is ``10``.
    do_project : bool, optional
        Reproject vertices onto the input surface when they are created or displaced.
        Default is ``True``.

    Returns
    -------
    list
        The vertices and faces of the new mesh.

    Notes
    -----
    This remeshing function only constrains the edges on the boundary of the mesh.
    To protect specific features or edges, please use :func:`remesh_constrained`.

    """
    V, F = mesh
    V = np.asarray(V, dtype=np.float64)
    F = np.asarray(F, dtype=np.int32)
    return meshing.remesh(V, F, target_edge_length, number_of_iterations)


def remesh_constrained(mesh: VerticesFaces,
                       target_edge_length: float,
                       protected_edges: List[Tuple[int, int]],
                       number_of_iterations: int = 10,
                       do_project: bool = True) -> VerticesFacesNumpy:
    """Remeshing of a triangle mesh.

    Parameters
    ----------
    mesh : tuple of vertices and faces
        The mesh to remesh.
    target_edge_length : float
        The target edge length.
    protected_edges : list
        A list of vertex pairs that identify protected edges of the mesh.
    number_of_iterations : int, optional
        Number of remeshing iterations.
        Default is ``10``.
    do_project : bool, optional
        Reproject vertices onto the input surface when they are created or displaced.
        Default is ``True``.

    Returns
    -------
    list
        The vertices and faces of the new mesh.

    """
    raise NotImplementedError



import numpy as np
from compas.plugins import plugin
from compas_cgal._cgal import booleans

from typing_extensions import Literal

from compas_cgal.types import VerticesFaces
from compas_cgal.types import VerticesFacesNumpy


__all__ = [
    'boolean_union',
    'boolean_difference',
    'boolean_intersection'
]


def _boolean(A: VerticesFaces,
             B: VerticesFaces,
             operation: Literal['union', 'difference', 'intersection']) -> VerticesFacesNumpy:
    """Wrapper for all boolean operations.

    Parameters
    ----------
    A : tuple of vertices and faces
        Mesh A.
    B : tuple of vertices and faces
        Mesh B.
    operation : {'union', 'difference', 'intersection'}
        The type of boolean operation.

    Returns
    -------
    list
        The vertices of the boolean mesh,
        and the faces of the boolean mesh,
        as a list of two numpy arrays.

    Raises
    ------
    NotImplementedError
        If the operation type is not supported.

    """
    VA, FA = A
    VB, FB = B
    VA = np.asarray(VA, dtype=np.float64)
    FA = np.asarray(FA, dtype=np.int32)
    VB = np.asarray(VB, dtype=np.float64)
    FB = np.asarray(FB, dtype=np.int32)

    if operation == 'union':
        result = booleans.boolean_union(VA, FA, VB, FB)
    elif operation == 'difference':
        result = booleans.boolean_difference(VA, FA, VB, FB)
    elif operation == 'intersection':
        result = booleans.boolean_intersection(VA, FA, VB, FB)
    else:
        raise NotImplementedError

    return result


@plugin(category='booleans', pluggable_name='boolean_union_mesh_mesh')
def boolean_union(A: VerticesFaces,
                  B: VerticesFaces) -> VerticesFacesNumpy:
    return _boolean(A, B, 'union')


@plugin(category='booleans', pluggable_name='boolean_difference_mesh_mesh')
def boolean_difference(A: VerticesFaces,
                       B: VerticesFaces) -> VerticesFacesNumpy:
    return _boolean(A, B, 'difference')


@plugin(category='booleans', pluggable_name='boolean_intersection_mesh_mesh')
def boolean_intersection(A: VerticesFaces,
                         B: VerticesFaces) -> VerticesFacesNumpy:
    return _boolean(A, B, 'intersection')



import numpy as np
from compas_cgal._cgal import measure
from compas.plugins import plugin

from compas_cgal.types import VerticesFaces


__all__ = [
    'volume'
]


@plugin(category='trimesh', pluggable_name='trimesh_volume')
def volume(mesh: VerticesFaces) -> float:
    """Compute the volume of a closed triangle mesh.

    Parameters
    ----------
    mesh : tuple of vertices and faces
        The mesh.

    Returns
    -------
    float
        The volume of the mesh.

    """
    V, F = mesh
    V = np.asarray(V, dtype=np.float64)
    F = np.asarray(F, dtype=np.int32)
    return measure.volume(V, F)



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Optional
from typing import List
from typing import Tuple
from nptyping import NDArray

import numpy as np
import scipy as sp
from compas.geometry import transform_points_numpy
from compas.geometry import Transformation
from compas.files import STL
from compas.files import OFF
from compas.files import PLY
from compas.datastructures import Mesh
from compas.numerical import connectivity_matrix
from compas.numerical import adjacency_matrix
from compas.numerical import degree_matrix
from compas.numerical import normrow

from compas_cgal.meshing import remesh


class TriMesh:
    """Class representing a triangle mesh with its vertices and faces stored as Numpy arrays.

    Parameters
    ----------
    vertices : array-like
        The vertices of the mesh.
    faces : array-like
        The faces of the mesh.

    Attributes
    ----------
    vertices : ndarray[(n, 3), float64]
        The vertices of the mesh stored as a `n` by 3 Numpy array of floats,
        with `n` the number of vertices.
    faces : ndarray[(f, 3), int32]
        The faces of the mesh stored as a `f` by 3 Numpy array of integers,
        with `f` the number of faces.
    edges : list of tuple(int, int) - readonly
        Edges of the mesh as pairs of vertices.
    adjacency : dict - readonly
        Vertex adjacency dict.
    C : sp.sparse.csr_matrix - readonly
        Sparse connectivity matrix.
    A : sp.sparse.csr_matrix - readonly
        Sparse adjacency matrix.
    D : sp.sparse.csr_matrix - readonly
        Sparse degree matrix.
    centroid : ndarray[(1, 3), float64] - readonly
        The centroid of the mesh.
    average_edge_length : float - readonly
        The average length og the edges.

    """

    def __init__(self, vertices, faces):
        self._vertices = None
        self._faces = None
        self._adjacency = None
        self._edges = None
        self.vertices = vertices
        self.faces = faces

    def __iter__(self):
        yield self.vertices
        yield self.faces

    def __getitem__(self, key):
        if key == 0:
            return self.vertices
        if key == 1:
            return self.faces
        raise KeyError

    def __setitem__(self, key, value):
        if key == 0:
            self.vertices = value
        elif key == 1:
            self.faces = value
        else:
            raise KeyError

    @property
    def vertices(self) -> NDArray[(Any, 3), np.float64]:
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        self._adjacency = None
        self._edges = None
        array = np.asarray(vertices, dtype=np.float64)
        rows, cols = array.shape
        if rows < 1 or cols != 3:
            raise Exception('The vertex input data is not valid.')
        self._vertices = array

    @property
    def faces(self) -> NDArray[(Any, 3), np.int32]:
        return self._faces

    @faces.setter
    def faces(self, faces):
        self._adjacency = None
        self._edges = None
        array = np.asarray(faces, dtype=np.int32)
        rows, cols = array.shape
        if rows < 1 or cols != 3:
            raise Exception('The face input data is not valid.')
        self._faces = array

    @property
    def adjacency(self) -> Dict[int, List[int]]:
        if not self._adjacency:
            adj = {i: [] for i in range(self.vertices.shape[0])}
            for a, b, c in self.faces:
                if b not in adj[a]:
                    adj[a].append(b)
                if c not in adj[a]:
                    adj[a].append(c)
                if a not in adj[b]:
                    adj[b].append(a)
                if c not in adj[b]:
                    adj[b].append(c)
                if a not in adj[c]:
                    adj[c].append(a)
                if b not in adj[c]:
                    adj[c].append(b)
            self._adjacency = adj
        return self._adjacency

    @property
    def edges(self) -> List[Tuple[int, int]]:
        if not self._edges:
            seen = set()
            edges = []
            for i, nbrs in self.adjacency.items():
                for j in nbrs:
                    if (i, j) in seen:
                        continue
                    seen.add((i, j))
                    seen.add((i, j))
                    edges.append((i, j))
            self._edges = edges
        return self._edges

    @property
    def C(self) -> sp.sparse.csr_matrix:
        return connectivity_matrix(self.edges, 'csr')

    @property
    def A(self) -> sp.sparse.csr_matrix:
        return adjacency_matrix(self.adjacency, 'csr')

    @property
    def D(self) -> sp.sparse.csr_matrix:
        return degree_matrix(self.adjacency, 'csr')

    @property
    def centroid(self):
        return np.mean(self.vertices, axis=0)

    @property
    def average_edge_length(self) -> float:
        return np.mean(normrow(self.C.dot(self.vertices)), axis=0)

    @classmethod
    def from_stl(cls: TriMesh, filepath: str, precision: Optional[str] = None) -> TriMesh:
        """Construct a triangle mesh from the data in an STL file."""
        stl = STL(filepath, precision)
        return cls(stl.parser.vertices, stl.parser.faces)

    @classmethod
    def from_ply(cls: TriMesh, filepath: str, precision: Optional[str] = None) -> TriMesh:
        """Construct a triangle mesh from the data in a PLY file."""
        ply = PLY(filepath, precision)
        return cls(ply.parser.vertices, ply.parser.faces)

    @classmethod
    def from_off(cls: TriMesh, filepath: str, precision: Optional[str] = None) -> TriMesh:
        """Construct a triangle mesh from the data in an OFF file."""
        off = OFF(filepath, precision)
        return cls(off.reader.vertices, off.reader.faces)

    @classmethod
    def from_mesh(cls: TriMesh, mesh: Mesh) -> TriMesh:
        """Construct a triangle mesh from a COMPAS mesh."""
        V, F = mesh.to_vertices_and_faces()
        return cls(V, F)

    def to_mesh(self) -> Mesh:
        """Convert the triangle mesh to a COMPAS mesh"""
        return Mesh.from_vertices_and_faces(self.vertices, self.faces)

    def copy(self) -> TriMesh:
        """Create an independent copy of the triangle mesh."""
        cls = type(self)
        return cls(self.vertices.copy(), self.faces.copy())

    def transform(self, T: Transformation):
        """Transform the triangle mesh."""
        self.vertices[:] = transform_points_numpy(self.vertices, T)

    def transformed(self, T: Transformation) -> TriMesh:
        """Create a transformed copy of the triangle mesh."""
        mesh = self.copy()
        mesh.transform(T)
        return mesh

    def cull_vertices(self):
        """Remove unused vertices."""
        indices = np.unique(self.faces)
        indexmap = {index: i for i, index in enumerate(indices)}
        self.vertices = self.vertices[indices]
        for face in self.faces:
            face[0] = indexmap[face[0]]
            face[1] = indexmap[face[1]]
            face[2] = indexmap[face[2]]

    def remesh(self, target_length: float, iterations: int = 10):
        """Remesh the triangle mesh."""
        V, F = remesh(self, target_length, iterations)
        self.vertices = V
        self.faces = F



"""
********************************************************************************
compas_cgal
********************************************************************************

.. currentmodule:: compas_cgal

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    trimesh.TriMesh

Functions
=========

Booleans
--------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    booleans.boolean_union
    booleans.boolean_difference
    booleans.boolean_intersection

Intersections
-------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    intersections.intersection_mesh_mesh

Measure
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    measure.volume

Meshing
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    meshing.remesh

Slicer
------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    slicer.slice_mesh

"""
from __future__ import print_function

import os


__author__ = ["tom van mele"]
__copyright__ = "Block Research Group - ETH Zurich"
__license__ = "MIT License"
__email__ = "van.mele@arch.ethz.ch"
__version__ = "0.1.2"

HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))


__all_plugins__ = [
    'compas_cgal.booleans',
    'compas_cgal.intersections',
    'compas_cgal.meshing',
    'compas_cgal.slicer',
]

__all__ = ["HOME", "DATA", "DOCS", "TEMP"]



import numpy as np
from compas_cgal._cgal import slicer
from compas.plugins import plugin

from compas_cgal.types import VerticesFaces
from compas_cgal.types import ListOfPlanes
from compas_cgal.types import ListOfPolylinesNumpy


__all__ = [
    'slice_mesh'
]


@plugin(category='trimesh', pluggable_name='trimesh_slice')
def slice_mesh(mesh: VerticesFaces,
               planes: ListOfPlanes) -> ListOfPolylinesNumpy:
    """Slice a mesh by a list of planes.

    Parameters
    ----------
    mesh : tuple of vertices and faces
        The mesh to slice.
    planes : list of (point, normal) tuples or compas.geometry.Plane
        The slicing planes.

    Returns
    -------
    list of arrays
        The points defining the slice polylines.

    """
    vertices, faces = mesh
    points, normals = zip(*planes)
    V = np.asarray(vertices, dtype=np.float64)
    F = np.asarray(faces, dtype=np.int32)
    P = np.array(points, dtype=np.float64)
    N = np.array(normals, dtype=np.float64)

    pointsets = slicer.slice_mesh(V, F, P, N)
    return pointsets



import numpy as np
from compas_cgal._cgal import intersections
from compas.plugins import plugin

from compas_cgal.types import VerticesFaces
from compas_cgal.types import ListOfPolylinesNumpy


__all__ = [
    'intersection_mesh_mesh'
]


@plugin(category='intersections')
def intersection_mesh_mesh(A: VerticesFaces, B: VerticesFaces) -> ListOfPolylinesNumpy:
    """Compute the intersection of tow meshes.

    Parameters
    ----------
    A : tuple of vertices and faces
        Mesh A.
    B : tuple of vertices and faces
        Mesh B.

    Returns
    -------
    list of arrays of points
        The intersection polylines as arrays of points.

    """
    VA, FA = A
    VB, FB = B
    VA = np.asarray(VA, dtype=np.float64)
    FA = np.asarray(FA, dtype=np.int32)
    VB = np.asarray(VB, dtype=np.float64)
    FB = np.asarray(FB, dtype=np.int32)

    pointsets = intersections.intersection_mesh_mesh(VA, FA, VB, FB)

    return pointsets
